import argparse
import functools
import pathlib

import numpy as np

from .._version import version

from . import definitions
from . import config

#: DryMass analysis output suffix (appended to data path)
OUTPUT_SUFFIX = "_dm"


@functools.lru_cache(maxsize=32)
def input_setting(path, section, key):
    """Ask the user for a configuration key"""
    cfg = config.ConfigFile(path)
    sec = cfg[section]
    description = definitions.config[section][key][2]
    if key not in sec or np.isnan(sec[key]):
        val = input("Please enter '{}' ({}): ".format(key, description))
        cfg.set_value("meta", key, val)


def main(path=None, req_meta=[]):
    """Main user dialog with optional "meta" kwargs required"""
    # get directories
    if path is None:
        path_in = parse().resolve()
    else:
        path_in = pathlib.Path(path).resolve()
    path_out = path_in.with_name(path_in.name + OUTPUT_SUFFIX)
    path_out.mkdir(exist_ok=True)
    # print directories (wrapped with functools so only executed once)
    print_info(path_in, path_out)
    # user input meta data keyword values
    for mm in sorted(req_meta):
        input_setting(path=path_out,
                      section="meta",
                      key=mm)
    return path_in, path_out


@functools.lru_cache(maxsize=32)  # cached to avoid multiple prints
def parse():
    """Obtain the input data set path by parsing the command line"""
    print("DryMass version {}".format(version))
    parser = argparse.ArgumentParser(description='DryMass QPI analysis.')
    parser.add_argument('path', metavar='path', nargs='+', type=str,
                        help='Data path')
    args = parser.parse_args()
    # Workaround: We use nargs='+' and join the input to support white
    # spaces in path names.
    jpath = " ".join(args.path)
    path_in = pathlib.Path(jpath).resolve()
    return path_in


@functools.lru_cache(maxsize=32)  # cached to avoid multiple prints
def print_info(path_in, path_out):
    """Print input and output paths"""
    print("Input:  {}".format(path_in))
    print("Output: {}".format(path_out))



import copy

import matplotlib as mpl
import matplotlib.pylab as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np


CM_PHASE = "viridis"
CM_PHASE_ERROR = copy.copy(plt.get_cmap("seismic"))
CM_PHASE_ERROR.set_over("g", 1)
CM_PHASE_ERROR.set_under("g", 1)
CM_INTENSITY = "gray"
CM_REFRACTIVE_INDEX = "gnuplot2"


try:
    mpl.rcParams['font.family'] = "sans-serif"
    mpl.rcParams["text.usetex"] = False
except TypeError:  # building the docs
    pass


def add_cbar(ax, mapper, cbformat="%.2f", label="",
             loc="right", size="5%", labelloc=None):
    """Add a colorbar to a plot"""
    if labelloc is None:
        labelloc = loc
    divider = make_axes_locatable(ax)
    cax = divider.append_axes(loc, size=size, pad=0.05)
    acbar = plt.colorbar(mapper, cax=cax, format=cbformat, label=label)
    acbar.ax.yaxis.set_ticks_position(labelloc)
    acbar.ax.yaxis.set_label_position(labelloc)
    return acbar


def plot_image(data, ax=None, imtype="phase", cbar=True, px_um=None,
               ret_cbar=False, **kwargs):
    """Plot an image

    Parameters
    ----------
    data: 2d np.ndarray
        Input image
    ax: matplotlib.Axes
        Axis to plot to
    imtype: str
        One of ["intensity", "phase", "phase_error",
        "refractive index"].
    cbar: bool
        Whether to add a colorbar.
    px_um: float
        Pixel size [µm]
    ret_cbar: bool
        Whether to return the colorbar.
    kwargs: dict
        Keyword arguments to `plt.imshow`.

    Returns
    -------
    ax [, cbar]:
        Axis and colorbar.
    """
    if ax is None:
        ax = plt.subplot(111)

    if imtype == "phase":
        cmap = CM_PHASE
        gridcolor = "w"
        cbformat = "%.1f"
        cblabel = "[rad]"
    elif imtype == "intensity":
        cmap = CM_INTENSITY
        gridcolor = "k"
        cbformat = "%.3f"
        cblabel = "[a.u.]"
        # Make sure gray is at 1 in the colormap
        if "vmin" in kwargs and "vmax" in kwargs:
            vmin = kwargs["vmin"]
            vmax = kwargs["vmax"]
            if vmin < 1 and vmax > 1:
                diff = max(1 - vmin, vmax - 1)
                kwargs["vmin"] = 1 - diff
                kwargs["vmax"] = 1 + diff
    elif imtype == "refractive index":
        cmap = CM_REFRACTIVE_INDEX
        gridcolor = "w"
        cbformat = "%.3f"
        cblabel = ""
    elif imtype == "phase error":
        cmap = CM_PHASE_ERROR
        gridcolor = "k"
        if "vmin" not in kwargs and "vmax" not in kwargs:
            vmax = np.max(np.abs(data))
            kwargs["vmax"] = vmax
            kwargs["vmin"] = -vmax
        cbformat = "%.3f"
        cblabel = "[rad]"
    else:
        raise ValueError("Unknown image type: {}".format(imtype))

    if px_um is None:
        shx, shy = np.array(data.shape)
        unit = "px"
    else:
        shx, shy = np.array(data.shape) * px_um
        unit = "µm"

    mapper = ax.imshow(data,
                       cmap=cmap,
                       extent=(0, shy, 0, shx),
                       interpolation="bilinear",
                       origin="lower",
                       **kwargs)

    ax.set_xlabel("x [{}]".format(unit))
    ax.set_ylabel("y [{}]".format(unit))
    ax.grid(color=gridcolor, lw="1", alpha=.1)

    retval = [ax]

    if cbar:
        acbar = add_cbar(ax=ax,
                         mapper=mapper,
                         cbformat=cbformat,
                         label=cblabel)

        if ret_cbar:
            retval.append(acbar)

    if len(retval) == 1:
        return retval[0]
    else:
        return retval


def plot_qpi_phase(qpi, rois=None, path=None):
    """Plot phase data"""
    fig = plt.figure(figsize=(6, 4))
    ax1 = plt.subplot(111, title="sensor phase image")
    px_um = qpi["pixel size"] * 1e6
    plot_image(data=qpi.pha,
               ax=ax1,
               imtype="phase",
               cbar=True,
               px_um=px_um)
    if rois:
        for roi in rois:
            slx, sly = roi[1]
            x0 = slx.start * px_um
            x1 = slx.stop * px_um
            y0 = sly.start * px_um
            y1 = sly.stop * px_um
            box = mpl.patches.Rectangle(xy=(y0, x0),
                                        width=y1 - y0,
                                        height=x1 - x0,
                                        facecolor="none",
                                        edgecolor="w",
                                        )
            ax1.add_patch(box)
            ax1.text(y0, x0, roi[0],
                     horizontalalignment="left",
                     verticalalignment="bottom",
                     color="w")
    plt.tight_layout()
    if path:
        fig.savefig(path)
        plt.close()
    else:
        return fig


def plot_qpi_sphere(qpi_real, qpi_sim, path=None, simtype="simulation"):
    """Plot QPI sphere analysis data"""
    fig = plt.figure(figsize=(9, 5))
    px_um = qpi_real["pixel size"] * 1e6
    radius_um = qpi_sim["sim radius"] * 1e6
    center = qpi_sim["sim center"]
    index = qpi_sim["sim index"]

    real_phase = qpi_real.pha
    kw_phase = {"px_um": px_um,
                "cbar": True,
                "imtype": "phase",
                "vmin": real_phase.min(),
                "vmax": real_phase.max(),
                }

    real_inten = qpi_real.amp**2
    kw_inten = {"px_um": px_um,
                "cbar": True,
                "imtype": "intensity",
                "vmin": real_inten.min(),
                "vmax": real_inten.max(),
                }
    # real phase
    ax1 = plt.subplot(
        231, title="data (phase)")
    plot_image(data=real_phase, ax=ax1, **kw_phase)

    # simulated phase
    ax2 = plt.subplot(232, title=simtype + " (phase)")
    plot_image(data=qpi_sim.pha, ax=ax2, **kw_phase)
    ax2.text(0.01, .99,
             "index: {:.5f}\n".format(index)
             + "radius: {:.3f}µm".format(radius_um),
             horizontalalignment="left",
             verticalalignment="top",
             color="w",
             transform=ax2.transAxes,
             )

    # phase difference
    ax3 = plt.subplot(233, title="phase difference")
    errmax = qpi_sim.pha.max() * .2
    plot_image(data=qpi_sim.pha - real_phase, ax=ax3,
               imtype="phase error", vmax=errmax, vmin=-errmax,
               px_um=px_um)

    # real intensity
    ax4 = plt.subplot(234, title="data (intensity)")
    plot_image(data=real_inten, ax=ax4, **kw_inten)

    # computed intensity
    ax5 = plt.subplot(235, title=simtype + " (intensity)")
    plot_image(data=qpi_sim.amp**2, ax=ax5, **kw_inten)

    # plot detected radius
    for ax in [ax1, ax2, ax4, ax5]:
        circ = mpl.patches.Circle(xy=((center[1] + .5) * px_um,
                                      (center[0] + .5) * px_um),
                                  radius=radius_um,
                                  facecolor="none",
                                  edgecolor="w",
                                  ls=(0, (3, 8)),
                                  lw=.5,
                                  )
        ax.add_patch(circ)

    # line plot through center
    ax6 = plt.subplot(236, title="phase line plot")
    x = np.arange(qpi_real.shape[1]) * px_um
    ax6.plot(x, qpi_sim.pha[int(center[0])], label=simtype)
    ax6.plot(x, qpi_real.pha[int(center[0])], label="data")
    ax6.set_xlabel("[µm]")
    ax6.legend(loc="center right")

    plt.tight_layout(rect=(0, 0, 1, .95))

    # add identifier
    fig.text(x=.5, y=.99, s=qpi_sim["identifier"],
             verticalalignment="top",
             horizontalalignment="center",
             fontsize=14)

    if path:
        fig.savefig(path)
        plt.close()
    else:
        return fig



from os import fspath
import pathlib
import warnings

import numpy as np
import qpimage
import qpsphere
from skimage.external import tifffile
import skimage.filters

from . import search, util
from .roi import ROIManager


#: Default background correction keyword arguments
BG_DEFAULT_KW = {"fit_offset": "mean",
                 "fit_profile": "tilt",
                 "border_perc": 5,
                 "border_px": 5,
                 }
#: Output ROI qpimage.QPSeries data
FILE_ROI_DATA_H5 = "roi_data.h5"
#: Output phase/amplitude TIFF data
FILE_ROI_DATA_TIF = "roi_data.tif"
#: Output slice locations
FILE_SLICES = "roi_slices.txt"


def _bg_correct(qpi, which_data, bg_kw={}, bg_mask_thresh=np.nan,
                bg_mask_sphere_kw={}):
    if bg_kw:
        if not np.isnan(bg_mask_thresh):  # binary threshold mask
            if which_data == "phase":
                image = qpi.pha
            else:
                image = qpi.amp
            mask1 = image2mask(image,
                               value_or_method=bg_mask_thresh)
        else:
            mask1 = None
        if not np.isnan(bg_mask_sphere_kw["radial_clearance"]):  # sphere mask
            mask2 = qpsphere.cnvnc.bg_phase_mask_for_qpi(
                qpi=qpi,
                r0=bg_mask_sphere_kw["r0"],
                method="edge",
                model="projection",
                edgekw=bg_mask_sphere_kw["edgekw"],
                imagekw={},
                radial_clearance=bg_mask_sphere_kw["radial_clearance"])
        else:
            mask2 = None
        # combine masks
        if mask1 is None and mask2 is None:
            mask = None
        elif mask1 is None:
            mask = mask2
        elif mask2 is None:
            mask = mask1
        else:
            mask = np.logical_and(mask1, mask2)
        # perforn actual bg correction
        qpi.compute_bg(which_data=which_data,
                       from_mask=mask,
                       **bg_kw)


def _extract_roi(h5in, h5out, slout, imout, size_m, size_var, max_ecc,
                 dist_border, pad_border, exclude_overlap,
                 bg_amp_kw, bg_amp_bin, bg_amp_mask_sphere_kw,
                 bg_pha_kw, bg_pha_bin, bg_pha_mask_sphere_kw,
                 search_enabled):
    # Determine ROI location
    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps:
        rmgr = ROIManager(qps.identifier)
        if search_enabled:
            for ii in range(len(qps)):
                qpi = qps[ii]
                # find objects
                slices = search.search_phase_objects(
                    qpi=qpi,
                    size_m=size_m,
                    size_var=size_var,
                    max_ecc=max_ecc,
                    dist_border=dist_border,
                    pad_border=pad_border,
                    exclude_overlap=exclude_overlap)
                for jj, sl in enumerate(slices):
                    slident = "{}.{}".format(qpi["identifier"], jj)
                    rmgr.add(roislice=sl, image_index=ii,
                             roi_index=jj, identifier=slident)
            rmgr.save(slout)
        else:
            rmgr.load(slout)

    # Extract ROI images
    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps, \
            qpimage.QPSeries(h5file=h5out, h5mode="w") as qps_roi, \
            tifffile.TiffWriter(fspath(imout), imagej=True) as tf:
        for ii in range(len(qps)):
            # image to analyze
            qpi = qps[ii]
            # available ROIs
            rois = rmgr.get_from_image_index(ii)
            for jj, (rid, sl) in enumerate(rois):
                # Extract the ROI
                qpisl = qpi.__getitem__(sl)
                # amplitude bg correction
                _bg_correct(qpi=qpisl,
                            which_data="amplitude",
                            bg_kw=bg_amp_kw,
                            bg_mask_thresh=bg_amp_bin,
                            bg_mask_sphere_kw=bg_amp_mask_sphere_kw)
                # phase bg correction
                _bg_correct(qpi=qpisl,
                            which_data="phase",
                            bg_kw=bg_pha_kw,
                            bg_mask_thresh=bg_pha_bin,
                            bg_mask_sphere_kw=bg_pha_mask_sphere_kw)
                slident = "{}.{}".format(qpi["identifier"], jj)
                if rid != slident:
                    # This might happen if the user does not know the
                    # image identifier and builds his own `FILE_SLICES`.
                    msg = "Mismatch of slice and QPImage identifiers: " \
                          + "{} vs {}!".format(rid, slident)
                    warnings.warn(msg)
                    # override `slident` with user identifier
                    slident = rid
                qps_roi.add_qpimage(qpisl, identifier=slident)

        if len(qps_roi):
            # Write TIF
            # determine largest image
            sxmax = np.max([qq.shape[0] for qq in qps_roi])
            symax = np.max([qq.shape[1] for qq in qps_roi])
            dummy = np.zeros((2, sxmax, symax), dtype=np.float32)
            for qpir in qps_roi:
                dummy[0, :, :] = 0
                dummy[1, :, :] = 1
                res = 1 / qpir["pixel size"] * 1e-6  # use µm
                sx, sy = qpir.shape
                dummy[0, :sx, :sy] = qpir.pha
                dummy[1, :sx, :sy] = qpir.amp
                tf.save(data=dummy, resolution=(res, res, None))
    return rmgr


def extract_roi(h5series, dir_out, size_m, size_var=.5, max_ecc=.7,
                dist_border=10, pad_border=40, exclude_overlap=30.,
                bg_amp_kw=BG_DEFAULT_KW, bg_amp_bin=np.nan,
                bg_amp_mask_radial_clearance=np.nan,
                bg_pha_kw=BG_DEFAULT_KW, bg_pha_bin=np.nan,
                bg_pha_mask_radial_clearance=np.nan,
                bg_sphere_edge_kw={}, search_enabled=True,
                ret_roimgr=False, ret_changed=False):
    """Extract ROIs from a qpimage.QPSeries hdf5 file

    Parameters
    ----------
    h5series: str
        Path of qpimage.QPSeries hdf5 file
    dir_out: str
        Path to output directory
    size_m: float
        Approximate diameter of the specimen [m]
    size_var: float
        Allowed variation relative to specimen size
    max_ecc: float
        Allowed maximal eccentricity of the specimen
    dist_border: int
        Minimum distance of objects to image border [px]
    pad_border: int
        Padding of object regions [px]
    exclude_overlap: float
        Allowed distance between two objects [px]
    bg_amp_kw: dict or None
        Amplitude image background correction keyword arguments
        (see :func:`qpimage.QPImage.compute_bg`), defaults
        to `BG_DEFAULT_KW`, set to `None` to disable correction
    bg_amp_bin: float or str
        The amplitude binary threshold value or the method for binary
        threshold determination; see :mod:`skimage.filters`
        `threshold_*` methods
    bg_amp_mask_radial_clearance: float
        If not NaN, use :func:`qpsphere.cnvnc.bg_phase_mask_for_qpi`
        to compute a mask image and use it for amplitude
        background correction.
    bg_pha_kw: dict or None
        Phase image background correction keyword arguments
        (see :func:`qpimage.QPImage.compute_bg`), defaults
        to `BG_DEFAULT_KW`, set to `None` to disable correction
    bg_pha_bin: float or str
        The phase binary threshold value or the method for binary
        threshold determination; see :mod:`skimage.filters`
        `threshold_*` methods
    bg_pha_mask_radial_clearance: float
        If not NaN, use :func:`qpsphere.cnvnc.bg_phase_mask_for_qpi`
        to compute a mask image and use it for phase
        background correction.
    search_enabled: bool
        If True, perform automated search for ROIs using the
        parameters above. If False, extract the ROIs from `FILE_SLICES`
        and only perform background correction using the `bg_*`
        parameters.
    ret_roimgr: bool
        Return the ROIManager instance of the found ROIs
    ret_changed: bool
        Return boolean indicating whether the ROI data on disk was
        created/updated (True) or whether previously created ROI
        data was used (False).

    Notes
    -----
    The output hdf5 file `dir_out/FILE_ROI_DATA_H5` is a
    :class:`qpimage.QPSeries` file with the keyword "identifier"
    consisting of two hashes: one from the relevant arguments
    to this method and one from the file `dir_out/FILE_SLICES`.
    This is to ensure that user-manipulated data is taken into
    account when deciding whether the ROIs should be re-computed
    after an initial run.
    """
    h5in = pathlib.Path(h5series)
    dout = pathlib.Path(dir_out)

    h5out = dout / FILE_ROI_DATA_H5
    imout = dout / FILE_ROI_DATA_TIF
    slout = dout / FILE_SLICES

    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps:
        cfgid = util.hash_object([qps,
                                  size_m,
                                  size_var,
                                  max_ecc,
                                  dist_border,
                                  pad_border,
                                  exclude_overlap,
                                  bg_amp_kw,
                                  bg_amp_bin,
                                  bg_amp_mask_radial_clearance,
                                  bg_pha_kw,
                                  bg_pha_bin,
                                  bg_pha_mask_radial_clearance,
                                  bg_sphere_edge_kw,
                                  search_enabled,
                                  ])

    # Determine whether we have to extract the ROIs
    if h5out.exists() and slout.exists():
        slid = util.hash_file(slout)
        identifier = "{}-{}".format(cfgid, slid)
        with qpimage.QPSeries(h5file=h5out, h5mode="r") as qpo:
            if qpo.identifier == identifier:
                create = False
            else:
                create = True
    else:
        create = True

    if create:
        bg_amp_mask_sphere_kw = {
            "r0": size_m / 2,
            "edgekw": bg_sphere_edge_kw,
            "radial_clearance": bg_amp_mask_radial_clearance}
        bg_pha_mask_sphere_kw = {
            "r0": size_m / 2,
            "edgekw": bg_sphere_edge_kw,
            "radial_clearance": bg_pha_mask_radial_clearance}

        rmgr = _extract_roi(
            h5in=h5in,
            h5out=h5out,
            slout=slout,
            imout=imout,
            size_m=size_m,
            size_var=size_var,
            max_ecc=max_ecc,
            dist_border=dist_border,
            pad_border=pad_border,
            exclude_overlap=exclude_overlap,
            bg_amp_kw=bg_amp_kw,
            bg_amp_bin=bg_amp_bin,
            bg_amp_mask_sphere_kw=bg_amp_mask_sphere_kw,
            bg_pha_kw=bg_pha_kw,
            bg_pha_bin=bg_pha_bin,
            bg_pha_mask_sphere_kw=bg_pha_mask_sphere_kw,
            search_enabled=search_enabled,
            )
        # manually set the identifier with the updated file slout
        slid = util.hash_file(slout)
        identifier = "{}-{}".format(cfgid, slid)
        with qpimage.QPSeries(h5file=h5out, h5mode="a") as qpo:
            qpo.h5.attrs["identifier"] = identifier
    else:
        with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps:
            rmgr = ROIManager(qps.identifier)
        rmgr.load(slout)

    ret = [h5out]
    if ret_roimgr:
        ret.append(rmgr)
    if ret_changed:
        ret.append(create)
    if len(ret) == 1:
        ret = ret[0]
    return ret


def image2mask(image, value_or_method):
    """Convert an image to a binary mask

    Parameters
    ----------
    image: 2d np.ndarray
        Input image
    value_or_method: float or str
        Either a threshold value or a string naming a
        filter method in :mod:`skimage.filters`.
    """
    if isinstance(value_or_method, str):
        method = getattr(skimage.filters, value_or_method)
        return image < method(image)
    else:
        return image < value_or_method



"""Comparison of relative and absolute dry mass

Relative dry mass is the dry mass computed relative to the
surrounding medium. If the refractive index of the surrounding
medium does not match that of the intracellular fluid
(approximately 1.335), then the relative dry mass underestimates
the actual dry mass. For a spherical cell, the absolute (corrected)
dry mass can be computed as described in the theory section on
:ref:`dry mass computation <section_theory_dry_mass>`.

This examples compares the relative dry mass
(:py:func:`drymass.ansphere.relative_dry_mass`) to the
absolute dry mass corrected for a spherical phase object
(:py:func:`drymass.ansphere.absolute_dry_mass_sphere`).
From simulated phase images (projection approach, wavelength 550nm)
of two cell-like spheres with a radius of 10µm and dry masses of
50pg (n≈1.337) and 250pg (n≈1.346), the absolute and relative dry
masses are computed with varying refractive index of the medium.

At the refractive index of phosphate buffered saline (PBS),
absolute and relative dry mass are equivalent.
As the refractive index of the medium increases,
the relative drymass decreases linearly (independent of dry mass),
underestimating the actual dry mass.
"""
from drymass.anasphere import absolute_dry_mass_sphere, relative_dry_mass
import matplotlib
import matplotlib.pylab as plt
import numpy as np
import qpsphere

# refraction increment
alpha = .18  # [mL/g]

# general simulation parameters
model = "projection"
wavelength = 500e-9  # [m]
pixel_size = 1.8e-7  # [m]
grid_size = (200, 200)  # [px]

# sphere parameters
radius = 10  # [µm]
center = (100, 100)  # [px]

dry_masses = [50, 250]  # [pg]
medium_indices = np.linspace(1.335, 1.34, 5)

qpi_pbs = {}
m_abs = {}
m_rel = {}
phase_data = np.zeros(grid_size, dtype=float)
for m in dry_masses:
    # initiate results list
    m_abs[m] = []
    m_rel[m] = []
    # compute refractive index from dry mass
    r_m = radius * 1e-6
    alpha_m3g = alpha * 1e-6
    m_g = m * 1e-12
    n = 1.335 + 3 * alpha_m3g * m_g / (4 * np.pi * (r_m**3))
    for medium_index in medium_indices:
        # generate example dataset
        qpi = qpsphere.simulate(radius=r_m,
                                sphere_index=n,
                                medium_index=medium_index,
                                wavelength=wavelength,
                                pixel_size=pixel_size,
                                model=model,
                                grid_size=grid_size,
                                center=center)
        # absolute dry mass
        ma = absolute_dry_mass_sphere(qpi=qpi,
                                      radius=r_m,
                                      center=center,
                                      alpha=alpha,
                                      rad_fact=1.2)
        m_abs[m].append(ma * 1e12)
        # relative dry mass
        mr = relative_dry_mass(qpi=qpi,
                               radius=r_m,
                               center=center,
                               alpha=alpha,
                               rad_fact=1.2)
        m_rel[m].append(mr * 1e12)
        if medium_index == 1.335:
            qpi_pbs[m] = qpi

# plot results
fig = plt.figure(figsize=(8, 4.5))
matplotlib.rcParams["image.interpolation"] = "bicubic"
# phase images
kw = {"vmax": qpi_pbs[dry_masses[1]].pha.max(),
      "vmin": qpi_pbs[dry_masses[1]].pha.min()}

ax1 = plt.subplot2grid((2, 3), (0, 2))
ax1.set_title("{}pg (in PBS)".format(dry_masses[0]))
ax1.axis("off")
map1 = ax1.imshow(qpi_pbs[dry_masses[0]].pha, **kw)

ax2 = plt.subplot2grid((2, 3), (1, 2))
ax2.set_title("{}pg (in PBS)".format(dry_masses[1]))
ax2.axis("off")
ax2.imshow(qpi_pbs[dry_masses[1]].pha, **kw)

# overview plot
ax3 = plt.subplot2grid((2, 3), (0, 0), colspan=2, rowspan=2)
ax3.set_xlabel("medium refractive index")
ax3.set_ylabel("computed dry mass [pg]")
for m, c in zip(dry_masses, ["blue", "green"]):
    ax3.plot(medium_indices, m_abs[m], ls="solid", color=c,
             label="{}pg absolute".format(m))
    ax3.plot(medium_indices, m_rel[m], ls="dashed", color=c,
             label="{}pg relative".format(m))
ax3.legend()
plt.colorbar(map1, ax=ax3, fraction=.048, pad=0.1,
             label="phase [rad]")

plt.tight_layout()
plt.subplots_adjust(wspace=.14)
plt.show()



import pathlib
import shutil
import tempfile

import numpy as np
import qpimage

from drymass.cli import cli_analyze_sphere, config, dialog
from drymass.anasphere import FILE_SPHERE_DATA, FILE_SPHERE_STAT


def setup_test_data(radius_px=30, size=200, pxsize=1e-6, medium_index=1.335,
                    wavelength=550e-9, num=1):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    pha = (r < radius_px) * 1.3
    amp = .5 + np.roll(pha, 10) / pha.max()
    qpi = qpimage.QPImage(data=(pha, amp),
                          which_data="phase,amplitude",
                          meta_data={"pixel size": pxsize,
                                     "medium index": medium_index,
                                     "wavelength": wavelength})
    path_in = tempfile.mktemp(suffix=".h5", prefix="drymass_test_cli_convert")
    path_in = pathlib.Path(path_in)
    with qpimage.QPSeries(h5file=path_in, h5mode="w", identifier="tt") as qps:
        for ii in range(num):
            qps.add_qpimage(qpi, identifier="image_{}".format(ii))
    # add drymass configuration file
    path_out = path_in.with_name(path_in.name + dialog.OUTPUT_SUFFIX)
    path_out.mkdir()
    cfg = config.ConfigFile(path_out)
    cfg.set_value(section="meta", key="pixel size um", value=pxsize*1e6)
    cfg.set_value(section="meta", key="wavelength nm", value=wavelength*1e9)
    cfg.set_value(section="meta", key="medium index", value=medium_index)
    cfg.set_value(section="specimen", key="size um",
                  value=radius_px*2*pxsize*1e6)
    return qpi, path_in, path_out


def test_base():
    _, path_in, path_out = setup_test_data(num=2)
    h5data = cli_analyze_sphere(path=path_in, ret_data=True)

    with qpimage.QPSeries(h5file=h5data, h5mode="r") as qps:
        assert len(qps) == 2

    # This might fail if the default kwargs for sphere analysis change
    assert h5data == path_out / FILE_SPHERE_DATA.format("edge", "projection")

    # check existence of files
    assert h5data.exists()
    assert h5data.stat().st_size > 10000

    pathsl = path_out / FILE_SPHERE_STAT.format("edge", "projection")
    assert pathsl.exists()
    assert len(pathsl.read_bytes()) > 100

    try:
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib

from . import definitions
from .._version import version

#: DryMass configuration file name
FILE_CONFIG = "drymass.cfg"


class ConfigFile(object):
    def __init__(self, path):
        """DryMass configuration file management

        Manage configuration file of an experimental data set
        with restrictions imposed by
        :data:`drymass.cli.definitions.config`.

        Parameters
        ----------
        path: str
            path to the configuration file
        """
        path = pathlib.Path(path).resolve()
        if path.is_dir():
            path = path / FILE_CONFIG
        path.touch()
        self.path = path

    def __getitem__(self, section):
        """Get a configuration section

        Parameters
        ----------
        section: str
            the configuration section

        Returns
        -------
        sectiondict: dict
            the configuration section dictionary
        """
        datadict = self._parse()
        if section in datadict:
            return datadict[section]
        elif section in definitions.config:
            # return default values
            secd = {}
            for kk in definitions.config[section]:
                secd[kk] = definitions.config[section][kk][0]
            # write result
            datadict[section] = secd
            self._write(datadict)
            return secd
        else:
            raise ValueError("Unknown section title: {}".format(section))

    def __setitem__(self, section, sectiondict):
        """Replace a section in the configuration file

        Parameters
        ----------
        section: str
            the section name
        sectiondict: dict
            the configuration dictionary

        Notes
        -----
        The previous content of the configuration section
        in the configuration file is replaced.
        """
        datadict = self._parse()
        for key in sectiondict:
            self._check_value(section, key, sectiondict[key])
        datadict[section] = sectiondict
        self._write(datadict)

    def _check_value(self, section, key, value):
        """Check if a section/key/value pair is valid

        Raises `ValueError` if this is not the case.
        Returns `None`.
        """
        if section not in definitions.config:
            raise ValueError("Unknown section title: {}".format(section))
        if key not in definitions.config[section]:
            raise ValueError("Unknown key: {}: {}".format(section, key))
        type_func = definitions.config[section][key][1]
        try:
            type_func(value)
        except BaseException:
            raise ValueError("Wrong dtype: {}: {}={}".format(section,
                                                             key, value))

    def _parse_compat(self, section, key, value):
        if section == "bg":
            # drymass < 0.1.3: API changed in qpimage 0.1.6
            if (key in ["amplitude profile", "phase profile"]
                    and value == "ramp"):
                value = "tilt"
        elif section == "roi":
            # drymass <= 0.1.5: keys were renamed to reflect pixel units
            if key in ["dist border", "exclude overlap", "pad border"]:
                key += " px"
        return key, value

    def _parse(self):
        """Return full documentation

        Returns
        -------
        datadict: dict of dicts
            Full configuration

        Notes
        -----
        If a configuration section in self.path is incomplete,
        then the defaults are be inserted and self.path is
        overridden.
        """
        with self.path.open() as fd:
            data = fd.readlines()
        outdict = {}
        for line in data:
            line = line.strip()
            if (line.startswith("#") or
                    len(line) == 0):
                pass
            elif line.startswith("["):
                sec = line.strip("[]")
                outdict[sec] = {}
            else:
                key, val = line.split("=")
                key = key.strip()
                val = val.strip()
                # backwards compatibility:
                key, val = self._parse_compat(sec, key, val)
                key_func = definitions.config[sec][key][1]
                try:
                    val = key_func(val)
                except ValueError as e:
                    # This means that the `key_func` has a problem with this
                    # value. To make things transparent, we append information
                    # to the error message.
                    e.args += ('Invalid value in "[{}]: {} = {}"'.format(
                               sec, key, val),)
                    raise
                outdict[sec][key] = val
        # Insert default variables where missing
        must_write = False
        for sec in outdict:
            for key in definitions.config[sec]:
                if key not in outdict[sec]:
                    outdict[sec][key] = definitions.config[sec][key][0]
                    must_write = True
        if must_write:
            # Update the configuration file
            self._write(outdict)
        return outdict

    def _write(self, datadict):
        """Write configuration dictionary

        Parameters
        ----------
        datadict: dict of dicts
            the full configuration

        Notes
        -----
        The configuration key values are converted to the correct
        dtype before writing using the definitions given in
        definitions.py.
        """
        keys = sorted(list(datadict.keys()))
        lines = ["# DryMass version {}".format(version),
                 "# Configuration file documented at: ",
                 "# https://drymass.readthedocs.io/en/stable/configuration"
                 + "_file.html",
                 ]
        for kk in keys:
            lines.append("")
            lines.append("[{}]".format(kk))
            subkeys = sorted(list(datadict[kk].keys()))
            for sk in subkeys:
                value = datadict[kk][sk]
                typefunc = definitions.config[kk][sk][1]
                lines.append("{} = {}".format(sk, typefunc(value)))
        for ii in range(len(lines)):
            lines[ii] += "\n"
        with self.path.open("w") as fd:
            fd.writelines(lines)

    def remove_section(self, section):
        """Remove a section from the configuration file"""
        datadict = self._parse()
        datadict.pop(section)
        self._write(datadict)

    def set_value(self, section, key, value):
        """Set a configuration key value

        Parameters
        ----------
        section: str
            the configuration section
        key: str
            the configuration key in `section`
        value:
            the configuration key value

        Notes
        -----
        Valid section and key names are defined in definitions.py
        """
        # load, update, and save
        sec = self[section]
        sec[key] = value
        self[section] = sec



import pathlib
import warnings


class ROIManagerWarning(UserWarning):
    """Used for unexpected keyword arguments."""
    pass


class ROIManager(object):
    def __init__(self, identifier=None):
        """Manage regions of interest (ROI) of an image series

        Parameters
        ----------
        identifier: str or None
            The identifier of the image series.
        """
        if not (isinstance(identifier, str) or identifier is None):
            raise ValueError("`identifier` must be `None` or a string!")
        self.identifier = identifier
        self.rois = []

    def __len__(self):
        return self.rois.__len__()

    def add(self, roislice, image_index, roi_index, identifier):
        """Add a ROI to ROIManager

        Parameters
        ----------
        roislice: (slice, slice)
            Two slices indicating the x- and y-axis limits.
        image_index: int
            The image index of the image series.
        roi_index: int
            The ROI index in image `image_index`
        identifier: str
            The ROI identifier. If `self.identifier` is not contained
            within `identifier`, a `ROIManagerWarning` will be issued.
        """
        # verify identifier
        if not isinstance(identifier, str):
            raise ValueError("`identifier` must be a string!")
        if self.identifier and self.identifier not in identifier:
            msg = "Identifier of ROIManager `{}` ".format(self.identifier) \
                  + "does not match that of QPSeries `{}`.".format(identifier)
            warnings.warn(msg, ROIManagerWarning)
        # verify roislice
        if not (isinstance(roislice, (tuple, list))
                and len(roislice) == 2
                and isinstance(roislice[0], slice)
                and isinstance(roislice[1], slice)):
            raise ValueError("`roislice` must be a tuple of two slices!")
        self.rois.append((identifier, image_index, roi_index, roislice))

    def get_from_image_index(self, image_index):
        rois = [[r[0], r[3]] for r in self.rois if r[1] == image_index]
        return rois

    def load(self, path):
        """Load ROIs from a text file"""
        path = pathlib.Path(path)
        with path.open(mode="r") as fd:
            lines = fd.readlines()
        # remove empty lines
        lines = [ll for ll in lines if ll.strip()]
        for ll in lines:
            ll = ll.strip().split("\t")
            sldata = "".join([c for c in ll[3] if c in ",0123456789"])
            sldata = sldata.replace(",,", ",").split(",")
            sldata = [int(s) for s in sldata if s]
            slice1 = slice(min(sldata[0], sldata[1]),
                           max(sldata[0], sldata[1]))
            slice2 = slice(min(sldata[2], sldata[3]),
                           max(sldata[2], sldata[3]))
            self.add(identifier=ll[0],
                     image_index=int(ll[1]),
                     roi_index=int(ll[2]),
                     roislice=(slice1, slice2)
                     )

    def save(self, path):
        """Save ROIs to a text file (`path` will be overridden)"""
        path = pathlib.Path(path)
        with path.open(mode="w") as fd:
            for roi in self.rois:
                roi = [str(r) for r in roi]
                fd.write("\t".join(roi) + "\r\n")



import numbers
from os import fspath
import pathlib

import numpy as np
import qpformat
import qpimage
from skimage.external import tifffile

#: Output qpimage.QPSeries sensor data
FILE_SENSOR_DATA_H5 = "sensor_data.h5"
#: Output phase/amplitude TIFF sensor data
FILE_SENSOR_DATA_TIF = "sensor_data.tif"


def convert(path_in, dir_out, meta_data={}, holo_kw={},
            bg_data_amp=None, bg_data_pha=None, write_tif=False,
            ret_dataset=False, ret_changed=False):
    """Convert experimental data to `qpimage.QPSeries` on disk

    Parameters
    ----------
    bg_data_amp, bg_data_pha: None, int, or path to file
        The background data for phase and amplitude. One of

        - `None`:
          No background data
        - `int`:
          Image index (starting at 0) of the input data set
          to use as background data
        - `str`, `pathlib.Path`:
          Path to a separate file that is used for background
          correction, relative to the directory in which `path_in`
          is located (`path_in.parent`).
    """
    path = pathlib.Path(path_in).resolve()
    dout = pathlib.Path(dir_out).resolve()

    h5out = dout / FILE_SENSOR_DATA_H5
    imout = dout / FILE_SENSOR_DATA_TIF

    ds = qpformat.load_data(path=path, meta_data=meta_data, holo_kw=holo_kw)

    if not (bg_data_amp is None and bg_data_pha is None):
        # Only set background of data set if there is
        # a background defined.
        bgamp = get_background(bg_data=bg_data_amp,
                               dataset=ds,
                               which="amplitude")
        bgpha = get_background(bg_data=bg_data_pha,
                               dataset=ds,
                               which="phase")
        bg_data = qpimage.QPImage(data=(bgpha, bgamp),
                                  which_data=("phase", "amplitude"))
        ds.set_bg(bg_data)

    if h5out.exists():
        with qpimage.QPSeries(h5file=h5out, h5mode="r") as qpsr:
            if (ds.identifier == qpsr.identifier and
                    len(ds) == len(qpsr)):
                # file has same identifier and same number of QPSeries
                create = False
            else:
                create = True
    else:
        create = True

    if create:
        # Write h5 data
        ds.saveh5(h5out)

    if write_tif and (create or not imout.exists()):
        # Also write tif data
        h5series2tif(h5in=h5out, tifout=imout)

    ret = [h5out]
    if ret_dataset:
        ret.append(ds)
    if ret_changed:
        ret.append(create)
    if len(ret) == 1:
        ret = ret[0]
    return ret


def get_background(bg_data, dataset, which="phase"):
    """Obtain the background data for a dataset

    Parameters
    ----------
    bg_data: None, int, str, or pathlib.Path
        Represents the background data:

        - None: no background data
        - int: image with this index in `dataset` is used
          for background correction
        - str, pathlib.Path: An external file will be used for background
          correction.
    dataset: qpformat.dataset.SeriesData
        The dataset for which the background data is collected.
        No background correction is performed! `dataset` is needed
        for integer `bg_data` and for path-based `bg_data`
        (because of meta data and hologram kwargs).

    Returns
    -------
    bg: 2d np.ndarray
        The background data.
    """
    if which not in ["phase", "amplitude"]:
        raise ValueError("`which` must be 'phase' or 'amplitude'!")
    if bg_data is None:
        bg = np.ones(dataset.get_qpimage(0).shape)
    elif isinstance(bg_data, numbers.Integral):
        if bg_data < 0 or bg_data > (len(dataset)-1):
            msg = "Background {} index must be between 0 and {}!"
            raise ValueError(msg.format(which, len(dataset)-1))
        # indexing in configuration file starts at 0
        if which == "phase":
            bg = dataset.get_qpimage(bg_data).pha
        else:
            bg = dataset.get_qpimage(bg_data).amp
    elif isinstance(bg_data, (str, pathlib.Path)):
        bgpath = pathlib.Path(bg_data)
        dsbg = qpformat.load_data(path=bgpath,
                                  meta_data=dataset.meta_data,
                                  holo_kw=dataset.holo_kw)
        if len(dsbg) != 1:
            msg = "Background correction with series data not implemented!"
            raise NotImplementedError(msg)
        else:
            if which == "phase":
                bg = dsbg.get_qpimage(0).pha
            else:
                bg = dsbg.get_qpimage(0).amp
    else:
        msg = "Unknown type for {} `bg_data`: {}".format(which, bg_data)
        raise ValueError(msg)
    return bg


def h5series2tif(h5in, tifout):
    """Convert a qpimage.QPSeries file to a phase/amplitude TIFF file"""
    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps, \
            tifffile.TiffWriter(fspath(tifout), imagej=True) as tf:
        for ii in range(len(qps)):
            qpi = qps[ii]
            res = 1 / qpi["pixel size"] * 1e-6  # use µm
            dshape = (1, qpi.shape[0], qpi.shape[1])
            dataa = np.array(qpi.amp, dtype=np.float32).reshape(*dshape)
            datap = np.array(qpi.pha, dtype=np.float32).reshape(*dshape)
            data = np.vstack((datap, dataa))
            tf.save(data=data, resolution=(res, res, None))



import tempfile
import shutil

import numpy as np

from drymass.cli import config


def test_basic():
    path = tempfile.mkdtemp(prefix="drymass_test_config_")
    cfg = config.ConfigFile(path=path)
    assert isinstance(cfg["bg"], dict)
    shutil.rmtree(path, ignore_errors=True)


def test_compat_013():
    path = tempfile.mkdtemp(prefix="drymass_test_config_")
    cfg = config.ConfigFile(path=path)
    # initialize config
    cfg.set_value("bg", "phase profile", "tilt")
    cfg.set_value("bg", "amplitude profile", "tilt")
    filepath = cfg.path
    # simulate old behavior
    data = filepath.open().read()
    data = data.replace(" profile = tilt\n", " profile = ramp\n")
    filepath.write_text(data)
    # check fix
    cfg2 = config.ConfigFile(path=path)
    assert cfg2["bg"]["phase profile"] == "tilt"
    assert cfg2["bg"]["amplitude profile"] == "tilt"
    shutil.rmtree(path, ignore_errors=True)


def test_compat_015():
    path = tempfile.mkdtemp(prefix="drymass_test_config_")
    cfg = config.ConfigFile(path=path)
    # initialize config
    cfg.set_value("roi", "dist border px", 3)
    cfg.set_value("roi", "exclude overlap px", 5)
    cfg.set_value("roi", "pad border px", 7)
    filepath = cfg.path
    # simulate old behavior
    data = filepath.open().read()
    data = data.replace("dist border px = ", "dist border = ")
    data = data.replace("exclude overlap px = ", "exclude overlap = ")
    data = data.replace("pad border px = ", "pad border = ")
    filepath.write_text(data)
    # check fix
    cfg2 = config.ConfigFile(path=path)
    assert cfg2["roi"]["dist border px"] == 3
    assert cfg2["roi"]["exclude overlap px"] == 5
    assert cfg2["roi"]["pad border px"] == 7
    shutil.rmtree(path, ignore_errors=True)


def test_dtype():
    path = tempfile.mkdtemp(prefix="drymass_test_config_")
    cfg = config.ConfigFile(path=path)

    cfg["specimen"] = {"size um": 11.3}
    assert np.allclose(cfg["specimen"]["size um"], 11.3)

    cfg.set_value(section="bg",
                  key="amplitude border px",
                  value=6.3)
    assert isinstance(cfg["bg"]["amplitude border px"], int)
    assert np.allclose(cfg["bg"]["amplitude border px"], 6)
    shutil.rmtree(path, ignore_errors=True)


def test_invalid_1():
    path = tempfile.mkdtemp(prefix="drymass_test_config_")
    cfg = config.ConfigFile(path=path)

    try:
        cfg["invalid section"]
    except ValueError:
        pass
    else:
        assert False, "invalid section should not work"

    try:
        cfg["invalid section"] = {"key": "value"}
    except ValueError:
        pass
    else:
        assert False, "invalid section should not work"

    try:
        cfg.set_value("invalid section", "key", "value")
    except ValueError:
        pass
    else:
        assert False, "invalid section should not work"

    try:
        cfg.set_value("bg", "invalid key", "value")
    except ValueError:
        pass
    else:
        assert False, "invalid key should not work"

    try:
        cfg.set_value("bg", "enabled", "invalid value")
    except ValueError:
        pass
    else:
        assert False, "invalid value should not work (bool expected)"

    shutil.rmtree(path, ignore_errors=True)


def test_invalid_2():
    path = tempfile.mkdtemp(prefix="drymass_test_config_")
    cfg = config.ConfigFile(path=path)

    cfg.path.write_text("[bg]\nenabled = bad value\n")
    try:
        cfg["bg"]
    except ValueError:
        pass
    else:
        assert False, "invalid value should not work (bool expected)"
    finally:
        cfg.path.write_text("")

    shutil.rmtree(path, ignore_errors=True)


def test_write_complete():
    path = tempfile.mkdtemp(prefix="drymass_test_config_")
    cfg = config.ConfigFile(path=path)

    cfg.path.write_text("[bg]\nenabled = True\n")
    # This triggers a completion of the configuration section [bg]
    assert cfg["bg"]["phase profile"] == "tilt"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import os
import tempfile
import shutil

import numpy as np
import pytest
import qpimage

import drymass


def setup_test_data(radius_px=30, size=200, pxsize=1e-6, medium_index=1.335,
                    wavelength=550e-9, num=1):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    image = (r < radius_px) * 1.3
    qpi = qpimage.QPImage(data=image,
                          which_data="phase",
                          meta_data={"pixel size": pxsize,
                                     "medium index": medium_index,
                                     "wavelength": wavelength})
    path = tempfile.mktemp(suffix=".h5", prefix="drymass_test_sphere")
    dout = tempfile.mkdtemp(prefix="drymass_test_sphere_")
    with qpimage.QPSeries(h5file=path) as qps:
        for ii in range(num):
            qps.add_qpimage(qpi, identifier="test_{}".format(ii))
    return qpi, path, dout


def test_basic():
    qpi, path, dout = setup_test_data(num=2)
    path_out = drymass.analyze_sphere(path, dir_out=dout)

    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qpso:
        assert qpso[0].shape == qpi.shape
        assert qpso[0]["wavelength"] == qpi["wavelength"]
        assert qpso[0]["pixel size"] == qpi["pixel size"]
        assert qpso[0]["medium index"] == qpi["medium index"]
        assert qpso[0]["identifier"].count("sim:projection")

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


@pytest.mark.filterwarnings('ignore::drymass.anasphere.'
                            + 'EdgeDetectionFailedWarning',
                            'ignore::RuntimeWarning')
def test_radius_exceeds_image_size_warning():
    pxsize = 1e-6
    size = 200
    _qpi, path, dout = setup_test_data(pxsize=pxsize, size=size)
    path_out = drymass.analyze_sphere(path, dir_out=dout,
                                      r0=size*pxsize*2)
    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qpso:
        assert np.all(qpso[0].amp == 1)
        assert np.all(qpso[0].pha == 0)

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_recompute_broken_output_path():
    _qpi, path, dout = setup_test_data()
    path_out = drymass.analyze_sphere(path, dir_out=dout)
    # break it
    with path_out.open("w") as fd:
        fd.write("This file is broken intentionally!")
    _p, changed = drymass.analyze_sphere(path, dir_out=dout, ret_changed=True)
    assert changed, "broken file should be recomputed"
    # delete it
    path_out.unlink()
    _p, changed = drymass.analyze_sphere(path, dir_out=dout, ret_changed=True)
    assert changed, "non-existent file should be recomputed"

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_recompute_edge_when_imagekw_changes():
    _qpi, path, dout = setup_test_data()
    drymass.analyze_sphere(path, dir_out=dout, imagekw={"stop_dc": .9})
    _p, changed = drymass.analyze_sphere(path, dir_out=dout,
                                         imagekw={"stop_dc": .9},
                                         ret_changed=True)
    assert not changed, "same arguments"
    _p, changed = drymass.analyze_sphere(path, dir_out=dout,
                                         imagekw={"stop_dc": .8},
                                         ret_changed=True)
    assert not changed, "different imagekw when method is edge"

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_recompute_edge_when_otherkw_changes():
    _qpi, path, dout = setup_test_data()
    drymass.analyze_sphere(path, dir_out=dout,
                           alpha=0.18,
                           edgekw={"clip_rmax": 1.1})
    _p, changed = drymass.analyze_sphere(path, dir_out=dout,
                                         alpha=0.19,
                                         edgekw={"clip_rmax": 1.1},
                                         ret_changed=True)
    assert changed, "change due to other alpha"
    _p, changed = drymass.analyze_sphere(path, dir_out=dout,
                                         alpha=0.19,
                                         edgekw={"clip_rmax": 1.2},
                                         ret_changed=True)
    assert changed, "change due to other edgekw"

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
